#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <jni.h>
#include <jvmpi.h>
#include <gtk/gtk.h>
#include <gtkutils.h>
#include <jmp-config.h>
#include <hash.h>
#include <jmp.h>
#include <ui.h>
#include <cls.h>
#include <jthread.h>
#include <filterconfig.h>
#include <eventconfig.h>
#include <classlist_menu.h>
#include <methodlist_menu.h>
#include <threads_window.h>
#include <comparators.h>
#include <dumper.h>
#include <heapgraph.h>
#include <about_dialog.h>
#include <timerstack.h>
#include <heap_dump.h>
#include <monitor_dialog.h>

/** state tokens. */
static int        init_done = 0;
static int        quit = 0;
static int        is_updating = 1;

/** id of last dump file... */
static int dump_id = -1;

/** Are we dumping on a timer basis? */
static gint  current_timer = 0; /* the timer identifier. */
static long  dumptimer = 0;     /* The countdown value (10)... */
static long  dumptime = 0;      /* Current value (1, 2, 3, 4,....) */

static GtkWidget *statusbar;
static GtkTooltips *tooltips;
static GtkWidget *JMPmain;

/** arrays that hold the methods/objects we show. */
static cls** classlist = NULL;
static jthread** threadslist = NULL;
static method** methodlist = NULL;
static int classes_count = 0;
static int threads_count = 0;
static int methods_count = 0;

/** A class to show total numbers in class info window. */
static cls* total = NULL;

/** The window of the current threads..  This is used in the
    timerstack locking functions to decide iff we need to lock. */
static threads_window *threadswin = NULL;

static GtkListStore *object_list = NULL;
static int object_list_size = 0;
static GtkWidget *class_statusbar;

static GtkListStore *method_list = NULL;
static int method_list_size = 0;
static GtkWidget *method_statusbar;

static int max_class_rows = 100;
static int max_method_rows = 100;

/** This is the timeout function called from gtk-main.
 *  Check if it is time to quit.
 */
gint run_updates (gpointer data) {
    if (quit) {
	if (usingUI ())
	    gtk_main_quit ();   /* time to shut down. */
	ui_shutdown_complete ();
	return FALSE;           /* stop timer. */
    }
    
    if (usingUI ()) {
	updateUI (get_classes (), get_methods ());
    }
    
    if (dumptimer > 0) {
	dumptime++;
	if (dumptime == dumptimer) {
	    run_data_dump ();
	    dumptime = 0;
	}
    }
    
    return TRUE;                /* keep timer */
}

/** Disable closing of window.
 */
static gint delete_event (GtkWidget *widget, GdkEvent  *event, 
			  gpointer   data) {
    return(TRUE);
}

static void destroy (GtkWidget *widget, gpointer data) {
    gtk_main_quit();
}

void run_dump (GtkMenuItem *menuitem, gpointer user_data) {
    run_data_dump ();
}

void run_garbage_collector (GtkMenuItem *menuitem, gpointer user_data) {
    set_status (_("Requesting garbage collection ..."));
    run_GC ();
}

void run_heap_dumper (GtkMenuItem *menuitem, gpointer user_data) {
    set_status (_("Getting heap dump"));
    run_heap_dump ();
    set_status (_("Heap dumped"));
}

void run_monitor_dumper (GtkMenuItem *menuitem, gpointer user_data) {
    hashtab* monitors;
    set_status (_("Getting monitor dump"));
    monitors = run_monitor_dump ();
    show_monitor_dialog (monitors);
    set_status (_("Monitors dumped"));
    jmphash_for_each ((jmphash_iter_f)free_monitor_info, monitors);    
    jmphash_free (monitors);
}

void run_reset_counter (GtkMenuItem *menuitem, gpointer user_data) {
    set_status (_("Resetting counters..."));
    reset_counters ();
    set_status (_("Counters reseted"));
}

void run_restore_counter (GtkMenuItem *menuitem, gpointer user_data) {
    set_status (_("Resetoring counters..."));
    restore_counters ();
    set_status (_("Counters restored"));
}

static void freeze_ui (GtkWidget* button) {
    is_updating = !is_updating;
    set_status (is_updating ? 
		_("UI updating continued") : 
		_("UI updating frozen"));
    gtk_label_parse_uline (GTK_LABEL (GTK_BIN (GTK_BUTTON (button))->child), 
			   is_updating ? _("_Freeze ui") : _("Un_freeze"));
}

static void show_threads () {
    set_status ("showing threads...");
    if (threadswin == NULL)
	threadswin = create_threads_window (get_threads ());
    else 
	update_threads_window (threadswin, get_threads ());
    timerstacks_set_need_locks (1);
    gtk_widget_show_all (GTK_WIDGET (threadswin->window));
}

static void about_jmp (GtkMenuItem *menuitem, gpointer user_data) {
    GtkWidget *about = create_about_dialog ();	
    gtk_widget_show_all (about);
}

/** Add a button to the button box.
 * @param label the text on the button
 * @param tooltip the tool tip for the button giving a more detailed description
 *  of the actions of the button.
 * @param hbuttonbox the button box to add the button to.
 * @param handler the callback for the button.
 */
static void add_button (char* label, char* tooltip, 
			GtkWidget* hbuttonbox, GCallback* handler) {
    GtkWidget* button;
    guint button_key;
    button = gtk_button_new_with_label ("");
    button_key = 
	gtk_label_parse_uline (GTK_LABEL (GTK_BIN (button)->child), label);
    gtk_tooltips_set_tip (tooltips, button, tooltip, NULL);
    gtk_container_add (GTK_CONTAINER (hbuttonbox), button);
    g_signal_connect (G_OBJECT (button), "clicked",
		      G_CALLBACK (handler), button);
}

/** Build the main window.
 *  Names of widgets is not totally nice though...
 */
GtkWidget* create_JMP (void) {
    GtkWidget *JMP;
    GtkWidget *vbox;
    GtkWidget *menubar;
    GtkWidget *hbuttonbox;
    GtkWidget *heapgraph;
    GtkItemFactoryEntry menu_items[] = {
	{ _("/_File"),                  NULL, NULL,                  0, "<Branch>" },
	{ _("/File/_Dump"),             NULL, run_dump,              0, NULL },
	{ _("/File/_Reset counters"),   NULL, run_reset_counter,     0, NULL },
	{ _("/File/Re_store counters"), NULL, run_restore_counter,   0, NULL },
	{ _("/File/System GC"),         NULL, run_garbage_collector, 0, NULL },
	{ _("/File/_Heapdump"),         NULL, run_heap_dumper,       0, NULL },
	{ _("/File/_Monitors"),         NULL, run_monitor_dumper,    0, NULL },
	{ _("/File/_Freeze UI"),        NULL, freeze_ui,             0, NULL },
	{ _("/File/Threads"),           NULL, show_threads,          0, NULL },
	{ _("/_Options"),               NULL, NULL,                  0, "<Branch>" },
	{ _("/Options/Filter"),         NULL, filter_edit_options,   0, NULL },
	{ _("/Options/Events"),         NULL, event_window,          0, NULL },
	{ _("/_Help"),                  NULL, NULL,                  0, "<LastBranch>" },
	{ _("/_Help/About"),            NULL, about_jmp,             0, NULL },
    };

    GtkItemFactory *item_factory;
    GtkAccelGroup *accel_group;
    gint nmenu_items = sizeof (menu_items) / sizeof (menu_items[0]);


    JMP = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_signal_connect (GTK_OBJECT (JMP), "delete_event",
			GTK_SIGNAL_FUNC (delete_event), NULL);

    gtk_window_set_title (GTK_WINDOW (JMP), _("Java Memory Profiler - Main"));
    vbox = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (JMP), vbox);
    
    /* build menues. */
    accel_group = gtk_accel_group_new ();
    item_factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>", 
					 accel_group);
    gtk_item_factory_create_items (item_factory, nmenu_items, menu_items, NULL);
    gtk_window_add_accel_group (GTK_WINDOW (JMP), accel_group);
    menubar = gtk_item_factory_get_widget (item_factory, "<main>");
    gtk_box_pack_start (GTK_BOX (vbox), menubar, FALSE, FALSE, 0);
    
    /* build the button bar. */
    hbuttonbox = gtk_hbutton_box_new ();
    gtk_hbutton_box_set_spacing_default (0);
    gtk_box_pack_start (GTK_BOX (vbox), hbuttonbox, TRUE, TRUE, 0);

    add_button (_("_Dump"), _("Dump data to a text file"), 
		hbuttonbox, (GCallback*)run_dump);
    add_button (_("_Reset"), _("Reset counters to zero"), 
		hbuttonbox, (GCallback*)run_reset_counter);
    add_button (_("Re_store"), _("Restore counters to the standard (full) values"), 
		hbuttonbox, (GCallback*)run_restore_counter);
    add_button (_("System._GC"), _("Run the java virtual machines garbage collector"),
		hbuttonbox, (GCallback*)run_garbage_collector);
    add_button (_("_Heapdump"), _("Show the current heap"), 
		hbuttonbox, (GCallback*)run_heap_dumper);
    add_button (_("_Monitors"), _("Show the current monitors"), 
		hbuttonbox, (GCallback*)run_monitor_dumper);
    add_button (_("_Freeze ui"), _("Stop updating values in the JMP windows"),
		hbuttonbox, (GCallback*)freeze_ui);
    add_button (_("Threads"), _("Show the current threads"),
		hbuttonbox, (GCallback*)show_threads);
 
    /* add a heap size graph */
    heapgraph = get_heap_graph ();
    gtk_box_pack_start (GTK_BOX (vbox), heapgraph, TRUE, TRUE, 0);   

    /* add status bar. */
    statusbar = gtk_statusbar_new ();
    gtk_box_pack_start (GTK_BOX (vbox), statusbar, FALSE, FALSE, 0);

    return JMP;
}

/** array of the class compare functions... */
static int ((*cls_comprs[])(const void* v1, const void* v2)) = { cls_compr_name, 
								 cls_compr_instance, 
								 cls_compr_max_instance, 
								 cls_compr_size,
								 cls_compr_instance_gc };

/** our current comparator */
static int (*cls_compr) (const void* v1, const void* v2) = cls_compr_size;

static void cls_column_clicked (GtkWidget *treeviewcolumn, gpointer user_data) {
    int column = (int)user_data;
    if (cls_comprs[column] != NULL) {
	if (cls_compr != cls_comprs[column]) {
	    cls_compr = cls_comprs[column];
	}
    } else {
	fprintf (stdout, "Sort order not yet implemented.\n");
    }
}

static void build_object_window () {
    GtkWidget *vbox;
    GtkWidget* scroller;
    GtkWidget* tree;
    GtkTreeSelection* select;

    GtkWidget* jmpwin = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_signal_connect (GTK_OBJECT (jmpwin), "delete_event",
			GTK_SIGNAL_FUNC (delete_event), NULL);
    gtk_signal_connect (GTK_OBJECT (jmpwin), "destroy",
			GTK_SIGNAL_FUNC (destroy), NULL);
    gtk_window_set_title (GTK_WINDOW (jmpwin), _("Java Memory Profiler - Objects"));
    scroller = gtk_scrolled_window_new (NULL, NULL);
    vbox = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (jmpwin), vbox);
    gtk_box_pack_start (GTK_BOX (vbox), scroller, TRUE, TRUE, 0);

    object_list = gtk_list_store_new (ON_COLUMNS, G_TYPE_STRING, G_TYPE_LONG, 
				      G_TYPE_LONG, G_TYPE_LONG, G_TYPE_LONG, 
				      G_TYPE_POINTER);
    
    tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (object_list));
    add_column (tree, _("Class"), OCLASS_COLUMN, (gpointer)0, cls_column_clicked, 200, 0);
    add_column (tree, _("Instances"), OINSTANCE_COLUMN, (gpointer)1, cls_column_clicked, 80, 1);
    add_column (tree, _("Max instances"), OINSTANCE_MAX_COLUMN, (gpointer)2, cls_column_clicked, 80, 1);
    add_column (tree, _("Size"), OSIZE_COLUMN, (gpointer)3, cls_column_clicked, 80, 1);
    add_column (tree, _("#GC"), OGC_COLUMN, (gpointer)4, cls_column_clicked, 80, 1);
    gtk_tree_view_set_headers_clickable (GTK_TREE_VIEW (tree), TRUE);
    gtk_container_add (GTK_CONTAINER (scroller), tree);
    select = gtk_tree_view_get_selection (GTK_TREE_VIEW (tree));
    g_signal_connect (G_OBJECT (select), "changed",
		      G_CALLBACK (olist_row_changed),
		      object_list);
    gtk_signal_connect (GTK_OBJECT(tree), "button_press_event",
			GTK_SIGNAL_FUNC (olist_button_handler), 
			NULL);

    class_statusbar = gtk_statusbar_new ();
    gtk_box_pack_start (GTK_BOX (vbox), class_statusbar, FALSE, FALSE, 0);

    gtk_widget_set_usize (jmpwin, 550, 200);
    gtk_widget_show_all (jmpwin);	
}

/** array of method compare functions... */
static int ((*method_comprs[])(const void* v1, const void* v2)) = { method_compr_class, 
								    method_compr_class, 
								    method_compr_time, 
								    method_compr_calls,
								    method_compr_hold_time,
								    method_compr_total_time,
								    method_compr_objects,
								    method_compr_bytes,
};


/** The current method comparator */
static int (*method_compr) (const void* v1, const void* v2) = method_compr_time;

static void method_column_clicked (GtkWidget *treeviewcolumn, gpointer user_data) {
    int column = (int)user_data;
    if (method_comprs[column] != NULL) {
	if (method_compr != method_comprs[column]) {
	    method_compr = method_comprs[column];
	}
    } else {
	fprintf (stdout, "Sort order not yet implemented.\n");
    }
}

static void build_method_window () {
    GtkWidget* scroller;
    GtkWidget *tree;
    GtkTreeSelection *select;
    GtkWidget *vbox;

    GtkWidget* jmpwin = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_signal_connect (GTK_OBJECT (jmpwin), "delete_event",
			GTK_SIGNAL_FUNC (delete_event), NULL);
    gtk_signal_connect (GTK_OBJECT (jmpwin), "destroy",
			GTK_SIGNAL_FUNC (destroy), NULL);
    gtk_window_set_title (GTK_WINDOW (jmpwin), _("Java Memory Profiler - Methods"));
    scroller = gtk_scrolled_window_new (NULL, NULL);
    vbox = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (jmpwin), vbox);
    gtk_box_pack_start (GTK_BOX (vbox), scroller, TRUE, TRUE, 0);
    
    method_list = gtk_list_store_new (MN_COLUMNS, G_TYPE_STRING, G_TYPE_STRING, 
				      G_TYPE_STRING, G_TYPE_LONG, G_TYPE_STRING,
				      G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT64, 
				      G_TYPE_POINTER);
    
    tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (method_list));
    add_column (tree, _("Class"), MCLASS_COLUMN, (gpointer)0, method_column_clicked, 200, 0);
    add_column (tree, _("Method"), MMETHOD_NAME_COLUMN, (gpointer)1, method_column_clicked, 200, 0);
    add_column (tree, _("secs"), MSECS_COLUMN, (gpointer)2, method_column_clicked, 80, 1);
    add_column (tree, _("calls"), MCALLS_COLUMN, (gpointer)3, method_column_clicked, 80, 1);
    add_column (tree, _("subs sec"), MSUBS_COLUN, (gpointer)4, method_column_clicked, 80, 1);
    add_column (tree, _("total"), MTOTAL_COLUMN, (gpointer)5, method_column_clicked, 80, 1);
    add_column (tree, _ ("objects"), MOBJECTS_COLUMN, (gpointer) 6, method_column_clicked, 80, 1);
    add_column (tree, _ ("bytes"), MBYTES_COLUMN, (gpointer) 7, method_column_clicked, 80, 1);
    gtk_tree_view_set_headers_clickable (GTK_TREE_VIEW (tree), TRUE);
    gtk_container_add (GTK_CONTAINER (scroller), tree);    

    select = gtk_tree_view_get_selection (GTK_TREE_VIEW (tree));
    g_signal_connect (G_OBJECT (select), "changed",
		      G_CALLBACK (mlist_row_changed),
		      method_list);
    gtk_signal_connect (GTK_OBJECT(tree), "button_press_event",
			GTK_SIGNAL_FUNC (mlist_button_handler), 
			NULL);

    method_statusbar = gtk_statusbar_new ();
    gtk_box_pack_start (GTK_BOX (vbox), method_statusbar, FALSE, FALSE, 0);

    gtk_widget_set_usize (jmpwin, 800, 200);
    gtk_widget_show_all (jmpwin);
}

static void setup_object_tracing () {
    if (total == NULL)
	total = cls_new (_("Total"), _("Total"), NULL, 0, 0, NULL, 0, NULL);
    if (object_list == NULL)
	build_object_window ();
}

static void setup_method_tracing () {
    if (method_list == NULL)
	build_method_window ();
}

/** The thread used to run gtk-main. 
 */
void gtkthread (void* data) {
    /* avoid gtk if we are not using any graphical ui */
    if (usingUI ()) {
	int argc = 1;
	char** argv;
	argv = malloc (sizeof (*argv));
	argv[0] = "jmp";
	gtk_init (&argc, &argv);
	if (tracing_objects ()) {
	    setup_object_tracing ();
	}
	if (tracing_methods ()) {
	    setup_method_tracing ();
	}
	
	tooltips = gtk_tooltips_new ();
	JMPmain = create_JMP ();
	gtk_widget_show_all (JMPmain);
	current_timer = gtk_timeout_add (1000, run_updates, NULL);
	init_done = 1;
	set_status (_("Ok"));
	gtk_main ();
    } else {
	while (!quit) {
	    sleep (1);
	    run_updates (NULL);
	}
    }
}

void set_ui_update_interval (int seconds) {
    if (usingUI () && current_timer) {
	gtk_timeout_remove (current_timer);
	current_timer = gtk_timeout_add (seconds * 1000, run_updates, NULL);
    }
}

/** Tell the ui system to shut itself down. */
void quit_ui (void) {
    quit = 1;
    if (total != NULL)
	cls_free (total);
    if (classlist)
	free (classlist);
    if (threadslist)
	free (threadslist);
    if (methodlist)
	free (methodlist);
    total = NULL;
    classlist = NULL;
    threadslist = NULL;
    methodlist = NULL;
}

/** Count the number of threads that are currently active */
static void count_threads (void* data) {
    threads_count++;
}

/* Add a thread to the array of threads. */
static void add_threads_row (void* data) {
    jthread* t = (jthread*)data;
    /* do tests ? */
    threadslist[threads_count++] = t;
}

static int (*jthread_compr) (const void* v1, const void* v2) = jthread_compr_name;

/** Count the number of classes that have more than zero instances. */
static void count_classes (void* data, cls_usage** cu) {
    cls* c = (cls*)data;
    /* fill in filtered class usage.. */
    if (cls_get_instances (c) != 0 && 
	cls_get_filtered (c)) {
	classes_count++;
	cu[0]->totalAlloced += c->usage.totalAlloced;
	cu[0]->totalInstances += c->usage.totalInstances;
	cu[0]->maximumInstances += c->usage.maximumInstances;
	cu[0]->totalGC += c->usage.totalGC;
    }
    /* fill in heap usage. */
    cu[1]->totalAlloced += c->usage.totalAlloced;
    cu[1]->totalInstances += c->usage.totalInstances;
    cu[1]->maximumInstances += c->usage.maximumInstances;
    cu[1]->totalGC += c->usage.totalGC;    
}

/** Add a row to the array of data. */
static void add_class_row (void* data) {
    cls* c = (cls*)data;    
    if (cls_get_instances (c) != 0 && 
	(c == total || cls_get_filtered (c)))
	classlist[classes_count++] = c;
}

/** Add a row to the gtk CList (update the ui). */
static void add_class_row_to_list (cls* c, int row, GtkTreeIter* iter) {
    int fast = 0;

    if (row < object_list_size) {
	cls* c2;
	gtk_tree_model_get (GTK_TREE_MODEL (object_list), iter, OOBJECT_COLUMN, &c2, -1);
	if (c == c2) {
	    if (!cls_check_modified (c)) {
		gtk_tree_model_iter_next (GTK_TREE_MODEL (object_list), iter);
		return;
	    }
	    fast = 1;
	}
    } else {
	gtk_list_store_append (object_list, iter);
    }
    
    if (fast) {
	gtk_list_store_set (object_list, iter,
			    OINSTANCE_COLUMN, cls_get_instances (c),
			    OINSTANCE_MAX_COLUMN, cls_get_max_instances (c),
			    OSIZE_COLUMN, cls_get_size (c),
			    OGC_COLUMN, cls_get_total_gc (c),
			    -1);
    } else {
	gtk_list_store_set (object_list, iter,
			    OCLASS_COLUMN, cls_get_name (c),
			    OINSTANCE_COLUMN, cls_get_instances (c), 
			    OINSTANCE_MAX_COLUMN, cls_get_max_instances (c),
			    OSIZE_COLUMN, cls_get_size (c),
			    OGC_COLUMN, cls_get_total_gc (c),
			    OOBJECT_COLUMN, c,
			    -1);
    }
    cls_set_modified (c, 0);
    gtk_tree_model_iter_next (GTK_TREE_MODEL (object_list), iter);
}

static void get_time_string (char* buf, size_t len, jlong val) {
    jlong jsec = val / 1000000000ll;
    jlong jnsec = (val % 1000000000ll) / 1000;
    snprintf (buf, len, "%lld.%06lld", jsec, jnsec);
}

/** Add a row to the gtk CList (update the ui). */
static void add_method_row_to_list (method* m, int row, GtkTreeIter* iter) {
    char sec[64];
    char hsec[64];
    char tsec[64];
    methodtime* mt;
    int fast = 0;

    if (row < method_list_size) {
	method* m2;
	gtk_tree_model_get (GTK_TREE_MODEL (method_list), iter, MMETHOD_COLUMN, &m2, -1);
	if (m == m2) {
	    if (!method_check_modified (m)) {
		gtk_tree_model_iter_next (GTK_TREE_MODEL (method_list), iter);
		return;
	    }
	    fast = 1;	    
	}
    } else {
	gtk_list_store_append (method_list, iter);
    }

    mt = method_get_time_used (m);
    get_time_string (sec, 64, mt->tv);
    get_time_string (hsec, 64, mt->tv_hold);
    get_time_string (tsec, 64, mt->tv + mt->tv_hold);
    if (fast) {
	gtk_list_store_set (method_list, iter,
			    MSECS_COLUMN, sec,
			    MCALLS_COLUMN, method_get_calls (m),
			    MSUBS_COLUN, hsec,
			    MTOTAL_COLUMN, tsec,
			    MOBJECTS_COLUMN, method_get_allocated_objects (m),
			    MBYTES_COLUMN, method_get_allocated_memory (m),
			    -1);
    } else {
	cls* c = method_get_owner (m);
	gtk_list_store_set (method_list, iter,
			    MCLASS_COLUMN, cls_get_name (c),
			    MMETHOD_NAME_COLUMN, method_get_method_jmpname (m),
			    MSECS_COLUMN, sec,
			    MCALLS_COLUMN, method_get_calls (m),
			    MSUBS_COLUN, hsec,
			    MTOTAL_COLUMN, tsec,
			    MMETHOD_COLUMN, m,
			    MOBJECTS_COLUMN, method_get_allocated_objects (m),
			    MBYTES_COLUMN, method_get_allocated_memory (m),
			    -1);
    }
    method_set_modified (m, 0);
    gtk_tree_model_iter_next (GTK_TREE_MODEL (method_list), iter);
}

/** Count the number of methodods that have used any time. */
static void count_methods (void* data) {
    method* m = (method*)data;
    if (!methodtime_is_zero (&m->time_used) &&
	cls_get_filtered (method_get_owner (m)))
	methods_count++;
}

/** Add a row to the array of data. */
static void add_methods_row (void* data) {
    method* m = (method*)data;
    if (!methodtime_is_zero (&m->time_used) && 
	cls_get_filtered (method_get_owner (m)))
	methodlist[methods_count++] = m;
}

static void clear_usage_data (cls_usage **cp) {
    cls_usage* cu = cp[0];
    cls_usage* hu = cp[1];
    
    cu->totalAlloced = 0;
    cu->totalInstances = 0;
    cu->maximumInstances = 0;
    cu->totalGC = 0;
    hu->totalAlloced = 0;
    hu->totalInstances = 0;
    hu->maximumInstances = 0;
    hu->totalGC = 0;
}

static void set_status_internal (GtkWidget *statusbar, char* text) {
    if (statusbar) {
	guint context_id = 1;        
	gtk_statusbar_pop (GTK_STATUSBAR (statusbar), context_id);
	gtk_statusbar_push (GTK_STATUSBAR (statusbar), context_id, text);
    }
}

/** Set the status text... */
void set_status (char* text) {
    set_status_internal (statusbar, text);
}

typedef void (*add_row_func)(void*, int, GtkTreeIter*);

/** updates a tree with new data. 
 */
static void update_tree (GtkListStore* list, int count, int max_rows, void** items, 
			 add_row_func add_row_to_list, int current_size) {
    GtkTreeIter   iter;
    int i;
    gtk_tree_model_get_iter_first (GTK_TREE_MODEL (list), &iter);
    for (i = 0; i < count && i < max_rows; i++) {
	add_row_to_list (items[i], i, &iter); 
    }
    if (i < current_size) {
	gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (list), 
				       &iter, NULL, i);
	for (; i < current_size; i++) {
	    /* this updates iter to point to next valid iter.. */
	    gtk_list_store_remove (list, &iter);  
	}
    }
}

void update_class_tree (hashtab* classes) {
    cls_usage cu, hu;
    cls_usage* cp[2] = {&cu, &hu};
    char buf[64];
    int oldcount = classes_count;    
    classes_count = 0;

    clear_usage_data (cp);
    setup_object_tracing ();
    jmphash_lock (classes);
    jmphash_for_each_with_arg ((jmphash_iter_fa)count_classes, classes, cp); 
    classes_count++;
    add_heap_size_value (hu.totalAlloced, cu.totalAlloced, current_heap_size ());
    total->usage = cu;
    cls_set_modified (total, 1);
    if (oldcount != classes_count) {
	classlist = realloc (classlist, classes_count * sizeof (cls*));
	/* TODO handle NULL here */
    }
    classes_count = 0;
    add_class_row (total);
    jmphash_for_each ((jmphash_iter_f)add_class_row, classes);
    jmphash_unlock (classes);
    qsort (classlist, classes_count, sizeof (cls*), cls_compr); 
    update_tree (object_list, classes_count, max_class_rows, (void**)classlist, 
		 (add_row_func)add_class_row_to_list, object_list_size);
    object_list_size = 
	max_class_rows < classes_count ? max_class_rows : classes_count;
    snprintf (buf, 64, _("Showing %d classes out of %d"), 
	      object_list_size, classes_count);
    set_status_internal (class_statusbar, buf);
}

void update_method_tree (hashtab* methods) {
    char buf[64];
    int oldcount = methods_count;

    setup_method_tracing ();
    /* update the method list window.. */
    methods_count = 0;
    jmphash_lock (methods);
    jmphash_for_each ((jmphash_iter_f)count_methods, methods);
    if (oldcount != methods_count) {
	methodlist = realloc (methodlist, methods_count * sizeof (method*));
	/* TODO handle NULL here.. */
    }
    methods_count = 0;
    jmphash_for_each ((jmphash_iter_f)add_methods_row, methods);
    jmphash_unlock (methods);
    qsort (methodlist, methods_count, sizeof (method*), method_compr);
    update_tree (method_list, methods_count, max_method_rows, (void**)methodlist, 
		 (add_row_func)add_method_row_to_list, method_list_size);
    method_list_size = 
	max_method_rows < methods_count ? max_method_rows : methods_count;
    snprintf (buf, 64, _("Showing %d methods out of %d"), 
	      method_list_size, methods_count);
    set_status_internal (method_statusbar, buf);

}

/** Update the statistics. */
void updateUI (hashtab* classes, hashtab* methods) {
    if (!init_done)
	return;

    if (!is_updating)
	return;

    if (!quit && tracing_objects ()) 
	update_class_tree (classes);

    if (!quit && tracing_methods ())
	update_method_tree (methods);
	
    if (threadswin != NULL && GTK_WIDGET_VISIBLE (threadswin->window))
	update_threads_window (threadswin, get_threads ());
}

/** Return the id of the dump file. */
int dump_data (char* dumpfile, hashtab* classes, hashtab* methods, hashtab* threads) {
    FILE* f;
    int i;
    int oldcount = classes_count;
    char filename[128];
    char status[128];
    cls_usage cu, hu;
    cls_usage* cp[2] = {&cu, &hu};
    snprintf (filename, 128, "%s-%d.txt", 
	      (dumpfile == NULL ? "jmp_dump" : dumpfile), ++dump_id);    
    snprintf (status, 128, _("dumped to: %s"), filename);
    set_status (status);
    f = fopen (filename, "w");
    if (f == NULL)
	return -1;

    fprintf (f, "Threads\n");
    fprintf (f, "thread\tgroup\tparent\tclass\tmethod\n");
    fprintf (f, "--------------------------------------------------------------\n");
    if (threads) {
	oldcount = threads_count;
	threads_count = 0;
	jmphash_for_each ((jmphash_iter_f)count_threads, threads);
	if (oldcount != threads_count) 
	    threadslist = realloc (threadslist, threads_count * sizeof(jthread*));
	threads_count = 0;
	jmphash_for_each ((jmphash_iter_f)add_threads_row, threads);
	qsort(threadslist, threads_count, sizeof(jthread*), jthread_compr);
	for (i = 0; i < threads_count; i++){
	    dump_jthread_row (threadslist[i], f);
	}
    } else {
	fprintf (stderr, "threads hash is NULL, wont dump it\n");
    }

    fprintf (f, "\n\n\nClassdump\n");
    fprintf (f, "class_name\t#instaces\tmax #instances\tsize\t#GC\n");
    fprintf (f, "--------------------------------------------------------------\n");
    if (classes) {
	classes_count = 0;
	
	clear_usage_data (cp);
	jmphash_for_each_with_arg ((jmphash_iter_fa)count_classes, classes, cp); 
	if (total == NULL) /* we should be synchronized so this wont give duplicates. */
	    total = cls_new (_("Total"), _("Total"), NULL, 0, 0, NULL, 0, NULL);	    
	total->usage = hu;
	cls_set_modified (total, 1);
	dump_class_row (total, f);
	total->usage = cu;
	cls_set_modified (total, 1);
	dump_class_row (total, f);
	if (oldcount != classes_count) 
	    classlist = realloc (classlist, classes_count * sizeof (cls*));
	classes_count = 0;
	jmphash_for_each ((jmphash_iter_f)add_class_row, classes); 
	qsort (classlist, classes_count, sizeof (cls*), cls_compr);
	for (i = 0; i < classes_count; i++) 
	    dump_class_row (classlist[i], f);	
    } else {
	fprintf (stderr, "classes hash is NULL, wont dump it\n");
    }

    fprintf (f, "\n\n\nMethods\n");
    fprintf (f, "class name\tmethod\tsec\tcalls\tsubsec\t#alloced instances\t#alloced bytes\n");
    fprintf (f, "--------------------------------------------------------------\n");
    if (methods) {
	oldcount = methods_count;    
	methods_count = 0;
	jmphash_for_each ((jmphash_iter_f)count_methods, methods);
	if (oldcount != methods_count) 
	    methodlist = realloc (methodlist, methods_count * sizeof (method*));
	methods_count = 0;
	jmphash_for_each ((jmphash_iter_f)add_methods_row, methods);
	qsort (methodlist, methods_count, sizeof (method*), method_compr);
	for (i = 0; i < methods_count; i++) 
	    dump_method_row (methodlist[i], f);
    } else {
	fprintf (stderr, "methods hash is NULL, wont dump it\n");	
    }

    fflush (f);
    fclose (f);
    return dump_id;
}

/** Set the dump timer... */
void set_dump_timer (int dt) {
    dumptimer = dt;
}

/** Get the current dump timer. */
int get_dump_timer () {
    return dumptimer;
}

/** Check if we have any ui-handling to do. */
int events_pending (void) {
    return gtk_events_pending ();
}

/** Update the ui-toolkit (gtk_main_iteration ()) */
int ui_iteration (void) {
    return gtk_main_iteration ();
}

/** check if a threads window is showing. This info is needed to 
 *  decide if we should lock the timer stacks.
 */
int is_threads_window_showing () {
    return threadswin != NULL && 
	GTK_WIDGET_VISIBLE (threadswin->window);
}

/** Show deadlock */
void show_deadlock (visited_threads* vt) {
   GtkWidget *dialog, *label;
   char buf[300];
   dialog = gtk_dialog_new_with_buttons (_("Deadlock detected"),
                                         GTK_WINDOW (JMPmain),
                                         GTK_DIALOG_DESTROY_WITH_PARENT,
                                         _("OK"),
                                         GTK_RESPONSE_NONE,
                                         NULL);

   label = gtk_label_new (_("Warning deadlock detected!\n"));  
   gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), label);
   while (vt != NULL) {
       if (vt->next) 
	   snprintf (buf, 300, _("%s holding %s (%p), is trying to enter %s (%p)"), 
		     jthread_get_thread_name (vt->mi->owner), 
		     vt->mi->name, vt->mi->id, 
		     vt->next->mi->name, vt->next->mi->id);
       else {
	   snprintf (buf, 300, _("%s holding %s (%p)"), 
		     jthread_get_thread_name (vt->mi->owner), 
		     vt->mi->name, vt->mi->id); 
       }
       vt = vt->next;
       label = gtk_label_new (buf);  
       gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), label);
   }    
   
   g_signal_connect_swapped (GTK_OBJECT (dialog), 
                             "response", 
                             G_CALLBACK (gtk_widget_destroy),
                             GTK_OBJECT (dialog));
   gtk_widget_show_all (dialog);
}

/* Emacs Local Variables: */
/* Emacs mode:C */
/* Emacs c-indentation-style:"gnu" */
/* Emacs c-hanging-braces-alist:((brace-list-open)(brace-entry-open)(defun-open after)(substatement-open after)(block-close . c-snug-do-while)(extern-lang-open after)) */
/* Emacs c-cleanup-list:(brace-else-brace brace-elseif-brace space-before-funcall) */
/* Emacs c-basic-offset:4 */
/* Emacs End: */
